#define TM_LINEAR 0
#define TM_REINHARDS 1
#define TM_REINHARDSMOD 2
#define TM_LOG 3
#define TM_ADAPTLOG 4

#include "ToneMapper.fxh"

static const float3 LUM_VECTOR = {0.3f, 0.58f, 0.12f};

cbuffer HDRParam
{
	float2 g_VPSize;
	float2 g_InvVPSize;
	
	bool  g_AutoKey;
	float g_Key;
	float g_DeltaTime;
	float g_BloomStrength;

	float g_BrightOffset;
	float g_BrightThreshold;

	float g_InvWhiteLumSQ;
	float g_AdaptLogB;

	int   g_BlurRadius;
	float g_BlurWeights[65];
};

Texture2DMS<float3, 4> g_txSceneAA;
Texture2D g_txScene;
Texture2D g_txBloom;

Texture2D g_txLuminance;
Texture2D g_txOldLuminance;

Texture2D g_txBloomInput;
Texture2D g_txLumInput;

SamplerState g_sBilinear
{
    Filter = MIN_MAG_LINEAR_MIP_POINT;
	AddressU = Clamp;
    AddressV = Clamp;
};

SamplerState g_sPoint
{
	Filter = MIN_MAG_MIP_POINT;
    AddressU = Clamp;
    AddressV = Clamp;
};

struct PS_IN
{
	float4 Position : SV_Position;
	float2 ScreenCoord : TEXCOORD0;
};

PS_IN VSQuad(in uint vid : SV_VertexID)
{
	PS_IN o;

	float tx = vid >> 1; //0, 0, 1, 1
	float ty = vid & 0x1; //0, 1, 0, 1

	float x = tx * 2.0f - 1.0f;
	float y = ty * 2.0f - 1.0f;
		
	o.Position = float4(x, y, 0, 1);
	o.ScreenCoord = float2(tx, 1 - ty);

	return o;
}

float GetAutoKey(float luminance)
{
	//return max(1.25f - 1/(luminance * 0.5f + 1), 0) + 0.12f; 
	return 0.1 + luminance*0.5;
	//return max(0.6f - 1 / (luminance * 0.5f + 1), 0) + 0.12f; 
	//return 0.28; 
}

float4 PSDownScale(PS_IN i, uniform bool aa) : SV_Target
{
	if (aa)
	{
		int2 texcrd = (i.ScreenCoord * g_VPSize);
		int2 scenecrd = texcrd << 1;

		//only load first sample
		float3 sum = 0;
		sum += g_txSceneAA.Load(scenecrd, 0, int2(0, 0));
		sum += g_txSceneAA.Load(scenecrd, 0, int2(0, 1));
		sum += g_txSceneAA.Load(scenecrd, 0, int2(1, 0));
		sum += g_txSceneAA.Load(scenecrd, 0, int2(0, 1));

		return float4(sum * 0.25f, 0);
	}
	else
	{
		//get the bilinear interpolated color from its 4 subpixels
		return float4(g_txScene.Sample(g_sBilinear, i.ScreenCoord).rgb, 0);
	}
}

float4 PSLumInit(PS_IN i) : SV_Target
{
	float LogLumSum = 0;

	for(int y = 0; y < 4; y++)
	{
		for(int x = 0; x < 4; x++)
		{
			// Compute the sum of log(luminance) throughout the sample points
			// a 8*64 sized input texture is assumed, because it wont be the exactly size (rather biger),
			// some pixel will be left out. however, it wont effect the final result much.
			
			//load pixels alredy bilnerar filtered to get as much info from the usually > 512x512 downscaled scene
			float2 tc = i.ScreenCoord + float2(x - 1.5f, y - 1.5f) / (4 * 64);
			LogLumSum += log(dot(g_txLumInput.Sample(g_sBilinear, tc), LUM_VECTOR) + 0.00001f);
		}
	}

	return LogLumSum / 16.0f;
}

float4 PSLumDown(PS_IN i) : SV_Target
{
	float sum = 0; 
    
	//build average of 8x8 area using 4x4 bilinear samples
	//target is a 8x8 texture
	for(int y = 0; y < 4; y++)
	{
		for(int x = 0; x < 4; x++)
		{
			// Compute the sum of luminance throughout the sample points
			float2 tc = i.ScreenCoord + float2(x - 1.5f, y - 1.5f) / (4 * 8);
			sum += g_txLumInput.Sample(g_sBilinear, tc).r;
		}
	}
        
    // Divide the sum to complete the average
    return sum / 16;
}

float4 PSLumFinal(PS_IN i, uniform bool adapt) : SV_Target
{
	float sum = 0; 
    
	for(int y = 0; y < 4; y++)
	{
		for(int x = 0; x < 4; x++)
		{
			// Compute the sum of luminance throughout the sample points
			float2 tc = i.ScreenCoord + float2(x - 1.5f, y - 1.5f) / 4;
			sum += g_txLumInput.Sample(g_sBilinear, tc).r;
		}
	}
        
    // Divide the sum to complete the average, and perform an exp() to complete
    // the average luminance calculation
	float newLum = exp(sum / 16);

	if (adapt)
	{
		float oldLum = g_txOldLuminance.Load(int3(0, 0, 0));

		float sigma = min(0.4f / (0.04f + newLum), 1);

		//interpolate tau from taurod and taucone depending on lum
		//Perceptual Effects in Real-time Tone Mapping: Equ(12)
		const float TauRod = 0.04f;
		const float TauCone = 0.01f;

		float Tau = (TauCone + (TauRod - TauCone) * sigma);

		//calculate adaption
		//Perceptual Effects in Real-time Tone Mapping: Equ(5)
		newLum = oldLum + (newLum - oldLum) * (1 - exp(-(g_DeltaTime*0.1) / Tau));
	}

	return newLum;
}

float3 BrightPassSample(float2 texcoord, float key, float luminance, bool tonemapped)
{
	float3 color = g_txBloomInput.Sample(g_sPoint, texcoord);

	// Determine what the pixel's value will be after tone-mapping occurs
	if (tonemapped)
		color *= key / (luminance + 0.00001f);
	
	// Subtract out dark pixels
	color -= g_BrightThreshold;
	
	// Clamp to 0
	color = max(color, 0.0f);
	
	// Map the resulting value into the 0 to 1 range. Higher values for
	// BRIGHT_PASS_OFFSET will isolate lights from illuminated scene 
	// objects.
	color /= (g_BrightOffset + color);

	return color;
}

float4 PSBloomHorizontal(PS_IN i, uniform bool tonemapped) : SV_Target
{
	//directly to bright pass
	float luminance = tonemapped ? g_txLuminance.Load(int3(0, 0, 0)) : 0;
	float key = tonemapped ? (g_AutoKey ? GetAutoKey(luminance) : g_Key) : 0;
	
	float3 sum = g_BlurWeights[0] * BrightPassSample(i.ScreenCoord, key, luminance, tonemapped);

	//sample left side
	for (int negx = 1; negx <= g_BlurRadius; negx++)
	{
		float2 tc = i.ScreenCoord;
		tc.x -= negx * g_InvVPSize.x;
		sum += g_BlurWeights[negx] * BrightPassSample(tc, key, luminance, tonemapped);
	}
	//sample right side
	for (int posx = 1; posx <= g_BlurRadius; posx++)
	{
		float2 tc = i.ScreenCoord;
		tc.x += posx * g_InvVPSize.x;
		sum += g_BlurWeights[posx] * BrightPassSample(tc, key, luminance, tonemapped);
	}

	return float4(sum, 0);
}

float4 PSBloomVertical(PS_IN i) : SV_Target
{
	float3 sum = g_BlurWeights[0] * g_txBloomInput.Sample(g_sPoint, i.ScreenCoord);

	//sample top half
	for (int negy = 1; negy <= g_BlurRadius; negy++)
	{
		float2 tc = i.ScreenCoord;
		tc.y -= negy * g_InvVPSize.y;
		sum += g_BlurWeights[negy] * g_txBloomInput.Sample(g_sPoint, tc);
	}
	//sample bottom half
	for (int posy = 1; posy <= g_BlurRadius; posy++)
	{
		float2 tc = i.ScreenCoord;
		tc.y += posy * g_InvVPSize.y;
		sum += g_BlurWeights[posy] * g_txBloomInput.Sample(g_sPoint, tc);
	}

	return float4(sum, 0);
}

float3 ToneMap(float3 color, uint tm, float key, float luminance)
{
	if (tm == TM_LINEAR)
		return TMLinear(color, key, luminance);
	else if (tm == TM_REINHARDS)
		return TMReinhard(color, key, luminance);
	else if (tm == TM_REINHARDSMOD)
		return TMReinhardMod(color, key, luminance, g_InvWhiteLumSQ);
	else if (tm == TM_LOG)
		return TMLogarithmic(color, key, luminance);
	else if (tm == TM_ADAPTLOG)
		return TMAdaptiveLog(color, key, luminance, g_AdaptLogB);

	return color;
}

float4 PSToneMapping(PS_IN i, uniform uint tm, uniform bool aa, uniform bool bloom) : SV_Target
{
	float luminance = g_txLuminance.Load(int3(0, 0, 0));
	float key = g_AutoKey ? GetAutoKey(luminance) : g_Key;

	int2 texcoord = i.ScreenCoord * g_VPSize;

	float3 color = 0;
	if (aa)
	{
		for (uint s = 0; s < 4; s++)
		{
			color += ToneMap(g_txSceneAA.Load(texcoord, s).rgb, tm, key, luminance);
		}

		color *= 0.25f;
	}
	else
	{
		color = ToneMap(g_txScene.Load(int3(texcoord, 0)).rgb, tm, key, luminance);
	}

	if (bloom)
	{
		color += g_BloomStrength * g_txBloom.Sample(g_sBilinear, i.ScreenCoord).rgb;
	}

	//return g_txLuminance.Sample(g_sBilinear, i.ScreenCoord).r;

	//if (i.ScreenCoord.y < 0.05)
	//{
	//	if (i.ScreenCoord.x < 0.05)
	//	{
	//		color = luminance;
	//	}
	//	else if (i.ScreenCoord.x < 0.1)
	//	{
	//		color = key;
	//	}
	//	else if (i.ScreenCoord.x < 0.15)
	//	{
	//		if (aa)
	//			color = ToneMap(g_txSceneAA.Load(texcoord, 0).rgb, TM_LINEAR, key, luminance);
	//		else 
	//			color = ToneMap(g_txScene.Load(int3(texcoord, 0)).rgb, TM_LINEAR, key, luminance);
	//	}
	//	else if (i.ScreenCoord.x < 0.2f)
	//	{
	//		color = 1;
	//	}
	//}

	return float4(color, 0);
}

float4 PSAddBloom(PS_IN i) : SV_Target
{
	float3 bloom = g_BloomStrength * g_txBloom.Sample(g_sBilinear, i.ScreenCoord).rgb;

	if (!any(bloom))
		discard;

	return float4(bloom, 0);
}

BlendState NoBlend
{
    AlphaToCoverageEnable = false;
    BlendEnable[0] = false;
	RenderTargetWriteMask[0] = 0x0F;
};

BlendState BlendAdd
{
    AlphaToCoverageEnable = false;
    BlendEnable[0] = true;
	RenderTargetWriteMask[0] = 0x0F;

	SrcBlend = ONE;
    DestBlend = ONE;
	BlendOp = ADD;
};

DepthStencilState NoDepth
{
	DepthEnable = false;
	DepthWriteMask = 0;
	StencilEnable = false;
};

RasterizerState NoCull
{
    CullMode = None;
    FillMode = Solid;
};

technique10 Render_DownScale
{
	pass P0
    {      
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDownScale(false)));
    }
}

technique10 Render_DownScaleAA
{
	pass P0
    {      
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSDownScale(true)));
    }
}

technique10 Render_LumInit
{
	pass P0
    {      
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSLumInit()));
    }
}

technique10 Render_LumDown
{
	pass P0
    {      
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSLumDown()));
    }
}

technique10 Render_LumFinal
{
	pass P0
    {      
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSLumFinal(false)));
    }
}

technique10 Render_LumFinal_Adapt
{
	pass P0
    {      
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSLumFinal(true)));
    }
}

technique10 Render_BloomHorizontal
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBloomHorizontal(false)));
	}
}

technique10 Render_BloomHorizontal_ToneMapped
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBloomHorizontal(true)));
	}
}

technique10 Render_BloomVertical
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBloomVertical()));
	}
}

technique10 Render_ToneMapping_Linear
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LINEAR, false, false)));
	}
}

technique10 Render_ToneMapping_Linear_AA
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LINEAR, true, false)));
	}
}

technique10 Render_ToneMapping_Reinhard
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDS, false, false)));
	}
}

technique10 Render_ToneMapping_Reinhard_AA
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDS, true, false)));
	}
}

technique10 Render_ToneMapping_ReinhardMod
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDSMOD, false, false)));
	}
}

technique10 Render_ToneMapping_ReinhardMod_AA
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDSMOD, true, false)));
	}
}

technique10 Render_ToneMapping_Logarithmic
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LOG, false, false)));
	}
}

technique10 Render_ToneMapping_Logarithmic_AA
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LOG, true, false)));
	}
}

technique10 Render_ToneMapping_AdaptiveLog
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_ADAPTLOG, false, false)));
	}
}

technique10 Render_ToneMapping_AdaptiveLog_AA
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_ADAPTLOG, true, false)));
	}
}

technique10 Render_ToneMapping_Linear_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LINEAR, false, true)));
	}
}

technique10 Render_ToneMapping_Linear_AA_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LINEAR, true, true)));
	}
}

technique10 Render_ToneMapping_Reinhard_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDS, false, true)));
	}
}

technique10 Render_ToneMapping_Reinhard_AA_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDS, true, true)));
	}
}

technique10 Render_ToneMapping_ReinhardMod_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDSMOD, false, true)));
	}
}

technique10 Render_ToneMapping_ReinhardMod_AA_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_REINHARDSMOD, true, true)));
	}
}

technique10 Render_ToneMapping_Logarithmic_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LOG, false, true)));
	}
}

technique10 Render_ToneMapping_Logarithmic_AA_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_LOG, true, true)));
	}
}

technique10 Render_ToneMapping_AdaptiveLog_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_ADAPTLOG, false, true)));
	}
}

technique10 Render_ToneMapping_AdaptiveLog_AA_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSToneMapping(TM_ADAPTLOG, true, true)));
	}
}

technique10 Render_ToneMapping_None_Bloom
{
	pass P0
	{
		SetDepthStencilState(NoDepth, 0);
		SetRasterizerState(NoCull);
		SetBlendState(BlendAdd, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSAddBloom()));
	}
}

