//	This effect does all shadow map rendering
//	It can be used directrly for all general materials.
//
//
//	In some cases it might be neccessary to overwrite certain functionality.
//
//
//	For a user specific position, eg. when displacement mapping should be used or discard the pixel shader with an alphamap,
//	first the user input struct has to be defined
//
//	struct MyVertexInput
//	{
//		float3 Pos : POSITION;
//		float  PosOffset : TEXCOORD0;
//		float2 Tex : TEXCOORD1;
//	};
//	#define SM_INPUT MyVertexInput
//
//
//	Then the user specific function that computes the position in object space has to be written and defined
//
//	float3 GetPosWithOffset(MyVertexInput i)
//	{
//		return i.Pos + i.PosOffset;
//	}
//	#define SM_GETPOSITION GetPosWithOffset
//
//
//	To discard by an alphamap, the texture coordinate has fist to be passed to the pixel shader.
//	The additional parameters have to be defined in a struct
//
//	struct MyExtraData
//	{
//		float2 tc : TEXCOORD1;
//	};
//	#define SM_EXTRADATA MyExtraData
//
//	This struct must be initialized, again a fucntion with a define is used for that
//
//	MyExtraData InitExtraData(MyVertexInput i)
//	{
//		MyExtraData ext;
//		ext.tc = i.tc;
//		return ext;
//	}
//	#define SM_INITVSEXTRADATA InitExtraData
//
//	Then write and define the discard function
//
//	void DiscardAlpha(MyExtraData ext)
//	{
//		if (ext.tc.x > 0.5 || ext.tc.y > 0.5) //do some texture read instead
//			discard;
//	}
//	#define SM_DISCARD DiscardAlpha

//implements new semantic parameters: CUBEVIEWPROJ, LIGHTPOS, LIGHTRANGE
shared cbuffer cbPerObject
{
	row_major matrix g_matWorld;
	row_major matrix g_matViewProj : VIEWPROJ;
	row_major matrix g_matLightCubeVP[6] : CUBEVIEWPROJ;

	//biasing stuff
	float3 g_LightPos : LIGHTPOS;
	float3 g_LightDir : LIGHTDIR;
	float2 g_ConstSlopeBias : BIAS;
	float  g_LightRange : LIGHTRANGE;
};

#ifndef SM_INPUT
//this struct can be overwritten if additional input is needed for the GetPosition() fuction
struct SM_INPUT
{
	float3 Pos : POSITION;
	float3 Normal : NORMAL;
};
#endif

#ifndef SM_GETPOSITION
//this function can be overwritten e.g. for displacement mapped surfaces
float3 SM_GETPOSITION(SM_INPUT i)
{
	return mul(float4(i.Pos, 1.0f), g_matWorld).xyz; //defaul behavior
}
#endif

#ifndef SM_GETNORMAL
//this function can be overwritten e.g. when normals are stored in a texture
float3 SM_GETNORMAL(SM_INPUT i)
{
	return normalize(mul(i.Normal, (float3x3)g_matWorld));
}
#endif

#ifndef SM_EXTRADATA
//this empty struct can be overwritte with extra data that can be used in a user specific function in the pixel shader
struct SM_EXTRADATA
{
};
#endif

#ifndef SM_DISCARD
#define SM_DISCARD //will make the compiler ignore SM_DISCARD(i);
#endif

//define an empty extra data handler, this handler has to be overwritte when a user specific struct is used
#ifndef SM_INITVSEXTRADATA
SM_EXTRADATA SM_INITVSEXTRADATA(SM_INPUT i)
{
	SM_EXTRADATA ext;
	return ext;
}
#endif

struct PS_DEPTH_IN
{
	float4 Pos : SV_Position;
	SM_EXTRADATA ext;
};

struct PS_DEPTHNORMAL_IN
{
	float4 Pos : SV_Position;
	SM_EXTRADATA ext;
	float3 Normal : TEXCOORD7;
};

struct PS_LINEAR_IN
{
	float4 Pos : SV_Position;
	SM_EXTRADATA ext;
	float2 ZW : TEXCOORD7;
};

struct PS_LINEARNORMAL_IN
{
	float4 Pos : SV_Position;
	SM_EXTRADATA ext;
	float2 ZW : TEXCOORD6;
	float3 Normal : TEXCOORD7;
};

//struct PS_DEPTHCUBE_IN
//{
//	float4 Pos : SV_Position;
//	float3 LightDir : TEXCOORD0;
//	SM_EXTRADATA ext;
//};

//used for points or spotlights
//has additinal bias scale depending on distance
PS_DEPTH_IN VSDepthPoint(SM_INPUT i)
{
	PS_DEPTH_IN o;

	float3 pos = SM_GETPOSITION(i);
	float3 normal = SM_GETNORMAL(i);

	float3 lightdir = pos - g_LightPos;
	float lightdist = length(lightdir);
	lightdir *= 1 / lightdist;
	
	float distscale = 1 + lightdist / g_LightRange;
	float slope = 1 - saturate(abs(dot(normal, lightdir)));
	pos += lightdir * (g_ConstSlopeBias.x + slope * g_ConstSlopeBias.y) * distscale;

	o.Pos = mul(float4(pos, 1), g_matViewProj);

	o.ext = SM_INITVSEXTRADATA(i);

	return o;
}

//used for directional lighting
//has bias simple worldspace constant + slopescale
PS_DEPTH_IN VSDepthUniform(SM_INPUT i)
{
	PS_DEPTH_IN o;

	float3 pos = SM_GETPOSITION(i);
	float3 normal = SM_GETNORMAL(i);
	
	float slope = 1 - saturate(abs(dot(normal, g_LightDir)));
	pos += g_LightDir * (g_ConstSlopeBias.x + slope * g_ConstSlopeBias.y);
	
	o.Pos = mul(float4(pos, 1), g_matViewProj);

	o.ext = SM_INITVSEXTRADATA(i);

	return o;
}
//used for lispsm directional lighting
//additional camdist bias scale
PS_DEPTH_IN VSDepthLisp(SM_INPUT i)
{
	PS_DEPTH_IN o;

	float3 pos = SM_GETPOSITION(i);
	float3 normal = SM_GETNORMAL(i);
	
	float camdist = length(pos.xyz - g_LightPos);
	float distscale = camdist * 10 / g_LightRange;
	float slope = 1 - saturate(abs(dot(normal, g_LightDir)));
	pos += g_LightDir * (g_ConstSlopeBias.x + slope * g_ConstSlopeBias.y) * distscale;
	
	o.Pos = mul(float4(pos, 1), g_matViewProj);

	o.ext = SM_INITVSEXTRADATA(i);

	return o;
}

//float4 PSBlack(PS_DEPTH_IN i, out float depth : SV_Depth) : SV_Target
//{ 
//	SM_DISCARD(i.ext);
//	depth = 1;
//	return 0;
//}

float4 PSBlack(PS_DEPTH_IN i) : SV_Target
{ 
	SM_DISCARD(i.ext);
	return 0;
}

//PS_DEPTHCUBE_IN VSDepthCube(SM_INPUT i)
//{
//	PS_DEPTHCUBE_IN o;
//
//	float3 pos = SM_GETPOSITION(i);
//
//	o.Pos = mul(float4(i.Pos, 1.0f), g_matWLightVP);
//
//	float3 WorldPos = (mul(float4(pos, 1.0f), g_matWorld)).xyz;
//	o.LightDir = WorldPos - g_LightPos;
//
//	o.ext = SM_INITVSEXTRADATA(i);
//
//	return o;
//}
//
//float4 PSDepthCube(PS_DEPTHCUBE_IN i) : SV_Target
//{ 
//	SM_DISCARD(i.ext);
//	return length(i.LightDir) / g_LightRadius + 0.005f;
//}

struct GS_DEPTHCUBEGS_IN
{
	float4 Pos : SV_Position;
	SM_EXTRADATA ext;
};

struct PS_DEPTHCUBEGS_IN
{
	float4 Pos		: SV_Position;
	uint RTIndex	: SV_RenderTargetArrayIndex;
	SM_EXTRADATA ext;
};

GS_DEPTHCUBEGS_IN VSDepthCubeGS(SM_INPUT i)
{
	GS_DEPTHCUBEGS_IN o;

	float3 pos = SM_GETPOSITION(i);
	float3 normal = SM_GETNORMAL(i);

	float3 lightdir = pos - g_LightPos;
	float lightdist = length(lightdir);
	lightdir *= 1 / lightdist;
	
	float distscale = lightdist * 10 / g_LightRange;
	float slope = saturate(abs(dot(normal, lightdir)));

	pos += lightdir * (g_ConstSlopeBias.x + slope * g_ConstSlopeBias.y) * distscale;

	o.Pos = float4(pos, 1);
	
	o.ext = SM_INITVSEXTRADATA(i);

	return o;
}

[maxvertexcount(18)]
void GSDepthCubeGS(triangle GS_DEPTHCUBEGS_IN i[3], inout TriangleStream<PS_DEPTHCUBEGS_IN> CubeMapStream)
{
    for (int f = 0; f < 6; ++f)
    {
        // Compute screen coordinates
        PS_DEPTHCUBEGS_IN o;
        o.RTIndex = f;
        for (int v = 0; v < 3; v++)
        {
            o.Pos = mul(i[v].Pos, transpose(g_matLightCubeVP[f]));
			o.ext = i[v].ext;
            CubeMapStream.Append(o);
        }
        CubeMapStream.RestartStrip();
    }
}

float4 PSDepthCubeGS(PS_DEPTHCUBEGS_IN i) : SV_Target
{ 
	SM_DISCARD(i.ext);
	return 0;
}

//PS_LINEAR_IN VSLinear(SM_INPUT i)
//{
//	PS_LINEAR_IN o;
//
//	float3 pos = SM_GETPOSITION(i);
//
//	o.Pos = mul(float4(pos, 1), g_matViewProj);
//
//	o.ext = SM_INITVSEXTRADATA(i);
//	o.ZW = o.Pos.zw;
//
//	return o;
//}
//
//PS_LINEARNORMAL_IN VSLinearNormal(SM_INPUT i)
//{
//	PS_LINEARNORMAL_IN o;
//
//	float3 pos = SM_GETPOSITION(i);
//	float3 normal = SM_GETNORMAL(i);
//
//	o.Pos = mul(float4(pos, 1), g_matViewProj);
//
//	o.ext = SM_INITVSEXTRADATA(i);
//	o.ZW = o.Pos.zw;
//	o.Normal = normal;
//
//	return o;
//}

//struct PS_DEPTHNORMAL_OUT
//{
//	float4 LinearDepth : SV_Target0;
//	float4 Normal : SV_Target1;
//};
//
//float4 PSLinear(PS_LINEAR_IN i) : SV_Target
//{
//	SM_DISCARD(i.ext);
//	return i.ZW.x / i.ZW.y;
//}
//
//PS_DEPTHNORMAL_OUT PSLinearNormal(PS_LINEARNORMAL_IN i)
//{
//	SM_DISCARD(i.ext);
//
//	PS_DEPTHNORMAL_OUT o;
//
//	o.LinearDepth = i.ZW.x / i.ZW.y;
//	o.Normal = float4(i.Normal, 0);
//
//	return o;
//}

PS_DEPTH_IN VSDepth(SM_INPUT i)
{
	PS_DEPTH_IN o;

	float3 pos = SM_GETPOSITION(i);

	o.Pos = mul(float4(pos, 1), g_matViewProj);

	o.ext = SM_INITVSEXTRADATA(i);

	return o;
}

PS_DEPTHNORMAL_IN VSDepthNormal(SM_INPUT i)
{
	PS_DEPTHNORMAL_IN o;

	float3 pos = SM_GETPOSITION(i);
	float3 normal = SM_GETNORMAL(i);

	o.Pos = mul(float4(pos, 1), g_matViewProj);

	o.ext = SM_INITVSEXTRADATA(i);
	o.Normal = normal;

	return o;
}

float4 PSNormal(PS_DEPTHNORMAL_IN i) : SV_Target
{
	SM_DISCARD(i.ext);

	return float4(normalize(i.Normal), 0);
}

DepthStencilState DSState
{
	DepthEnable = true;
	DepthWriteMask = ALL;
	DepthFunc = LESS;
	StencilEnable = false;
};

BlendState NoBlendNoColorWrite
{
    BlendEnable[0] = FALSE;
	RenderTargetWriteMask[0] = 0;
	RenderTargetWriteMask[1] = 0xF;
	RenderTargetWriteMask[2] = 0xF;
	RenderTargetWriteMask[3] = 0xF;
	RenderTargetWriteMask[4] = 0xF;
	RenderTargetWriteMask[5] = 0xF;
};

BlendState NoBlend
{
	BlendEnable[0] = FALSE;
	RenderTargetWriteMask[0] = 0xF;
};

technique10 Render_DepthPoint
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetBlendState(NoBlendNoColorWrite, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSDepthPoint()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBlack()));
    }  
}

technique10 Render_DepthUniform
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetBlendState(NoBlendNoColorWrite, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSDepthUniform()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBlack()));
    }  
}

technique10 Render_DepthLisp
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetBlendState(NoBlendNoColorWrite, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSDepthLisp()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBlack()));
    }  
}

//technique10 Render_DepthCube
//{
//    pass p0
//    {
//		SetDepthStencilState(DSState, 0);
//		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
//    
//        SetVertexShader(CompileShader(vs_4_0, VSDepthCube()));
//        SetGeometryShader(NULL);
//        SetPixelShader(CompileShader(ps_4_0, PSDepthCube()));
//    }  
//}

technique10 Render_DepthCubeGS
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetBlendState(NoBlendNoColorWrite, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSDepthCubeGS()));
        SetGeometryShader(CompileShader(gs_4_0, GSDepthCubeGS()));
        SetPixelShader(CompileShader(ps_4_0, PSDepthCubeGS()));
    }  
}

technique10 Render_Depth
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetBlendState(NoBlendNoColorWrite, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSDepth()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBlack()));
    }  
}

technique10 Render_DepthNormal
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSDepthNormal()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSNormal()));
    }  
}