cbuffer cb
{
	uint g_PatchSize;
	float g_PatchScale;

	//modify stuff
	float2 g_Center;
	float g_Radius;
	float g_Falloff;
	float g_Strength;
	float g_SmoothTarget;

	//quad position
	int3 g_QuadPos; //position and size of quad
	float3 g_QuadRect; //position and size of quad to read
};

Texture2D g_txHeightmap;

SamplerState g_sLinearFilter
{
    Filter = MIN_MAG_LINEAR_MIP_POINT; 
    AddressU = Clamp;
    AddressV = Clamp;
};

struct PS_IN
{
	float4 pos : SV_Position;
	float2 tc : TEXCOORD0;
};

PS_IN VSFullScreenQuad(in uint vid : SV_VertexID)
{
	PS_IN o;

	float tx = (vid / 2);
	float ty = (vid % 2); 

	float x = tx * 2.0f - 1.0f;
	float y = ty * 2.0f - 1.0f;
		
	o.pos = float4(x, y, 0, 1);
	o.tc = float2(tx, 1 - ty);

	return o;
}

PS_IN VSQuad(in uint vid : SV_VertexID)
{
	PS_IN o;

	float invrtsize = 1.0f / (g_PatchSize + 3);

	float2 p = g_QuadPos.xy * invrtsize;
	float2 tc = g_QuadRect.xy;

	if (vid > 1)
	{
		p.x += g_QuadPos.z * invrtsize;
		tc.x += g_QuadRect.z;
	}

	if (vid & 0x1)
	{
		p.y += g_QuadPos.z * invrtsize;
		tc.y += g_QuadRect.z;
	}
		
	o.pos = float4(p * float2(2, -2) + float2(-1, 1), 0, 1);
	o.tc = float2(tc.x, tc.y);

	return o;
}

float4 PSBuildNormals(PS_IN i) : SV_Target
{
	int2 nxy = i.tc * (g_PatchSize + 1); //coordinates of which normal is build
	
	float hy0 = g_txHeightmap.Load(int3(nxy.x + 1, nxy.y + 0, 0)).r;
	float hy1 = g_txHeightmap.Load(int3(nxy.x + 1, nxy.y + 2, 0)).r;
	//float hc = g_txHeightmap.Load(int3(nxy.x + 1, nxy.y + 1, 0)).r;
	float hx0 = g_txHeightmap.Load(int3(nxy.x + 0, nxy.y + 1, 0)).r;
	float hx1 = g_txHeightmap.Load(int3(nxy.x + 2, nxy.y + 1, 0)).r;

	//float2 nx0 = normalize(float2(hc - hx0, g_PatchScale));
	//float2 nx1 = normalize(float2(hx1 - hc, g_PatchScale));
	//float2 ny0 = normalize(float2(hc - hy0, g_PatchScale));
	//float2 ny1 = normalize(float2(hy1 - hc, g_PatchScale));

	//float3 n = float3(-nx0.x - nx1.x, -ny0.x - ny1.x, g_PatchScale);

	float3 n = float3((hx0 - hx1), (hy0 - hy1), 2 * g_PatchScale);

	return float4(normalize(n), 0);
}

float4 PSModifyHeight(PS_IN i) : SV_Target
{
	float2 coord = i.tc * (g_PatchSize + 3) - 1.5;
	float d = length(coord - g_Center);

	float val = g_Strength;

	if (d >= g_Radius + g_Falloff)
	{
		discard;
	}
	else if (d > g_Radius)
	{
		val *= (1 - (d - g_Radius) / g_Falloff);
	}

	return val;
}

float4 PSSmoothHeight(PS_IN i) : SV_Target
{
	float2 coord = i.tc * (g_PatchSize + 3) - 1.5;
	float d = length(coord - g_Center);

	float val = g_Strength;

	if (d >= g_Radius + g_Falloff)
	{
		discard;
	}
	else if (d > g_Radius)
	{
		val *= (1 - (d - g_Radius) / g_Falloff);
	}

	return float4(g_SmoothTarget, 0, 0, val);
}

float4 PSUpdateHeightmapLOD(PS_IN i) : SV_Target
{
	int2 coord = i.tc * (g_PatchSize + 1); //index to sample 

	float filtered = 0;

	const float weights[] = { 1 / 16.0f, 2 / 16.0f, 1 / 16.0f,
							  2 / 16.0f, 4 / 16.0f, 2 / 16.0f,
							  1 / 16.0f, 2 / 16.0f, 1 / 16.0f};

	//for (uint y = 0; y < 3; y++)
	//	for (uint x = 0; x < 3; x++)
	//	{
	//		float h = g_txHeightmap.Load(int3(coord + int2(x, y), 0)).r;
	//		filtered += h * weights[x + y * 3];
	//	}

	filtered = g_txHeightmap.Load(int3(coord + int2(1, 1), 0)).r;

	return filtered;
	//return max(0, min(1, i.tc.x)) * 10;
	//return 1;
}

RasterizerState RasNoCull
{
    CullMode = None;
    FillMode = Solid;
};

DepthStencilState DSState
{
	DepthEnable = false;
	StencilEnable = false;
};

BlendState NoBlend
{
    BlendEnable[0] = false;
	RenderTargetWriteMask[0] = 0x0f;
};

BlendState BlendAdd
{
	AlphaToCoverageEnable = false;
    BlendEnable[0] = true;
    BlendOp = ADD;
    SrcBlend = ONE;
    DestBlend = ONE;
	RenderTargetWriteMask[0] = 0x0f;
};

BlendState BlendAlpha
{
	AlphaToCoverageEnable = false;
    BlendEnable[0] = true;
    BlendOp = ADD;
    SrcBlend = SRC_ALPHA;
    DestBlend = INV_SRC_ALPHA;
	RenderTargetWriteMask[0] = 0x0f;
};

technique10 RenderNormals
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetRasterizerState(RasNoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSFullScreenQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSBuildNormals()));
    }  
}

technique10 ModifyHeight
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetRasterizerState(RasNoCull);
		SetBlendState(BlendAdd, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSFullScreenQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSModifyHeight()));
    }  
}

technique10 SmoothHeight
{
    pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetRasterizerState(RasNoCull);
		SetBlendState(BlendAlpha, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSFullScreenQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSSmoothHeight()));
    }  
}

technique10 UpdateHeightmapLOD
{
	pass p0
    {
		SetDepthStencilState(DSState, 0);
		SetRasterizerState(RasNoCull);
		SetBlendState(NoBlend, float4(0.0f, 0.0f, 0.0f, 0.0f), 0xffffffff);
    
        SetVertexShader(CompileShader(vs_4_0, VSQuad()));
        SetGeometryShader(NULL);
        SetPixelShader(CompileShader(ps_4_0, PSUpdateHeightmapLOD()));
    } 
}